# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.

import pandas as pd

from collections import OrderedDict
from datetime import datetime
from pathlib import Path

from nsys_recipe import log
from nsys_recipe.data_service import StatsService
from nsys_recipe.lib import helpers, recipe
from nsys_recipe.lib.args import Option
from nsys_recipe.log import logger


class CudaGpuMemTimeSum(recipe.Recipe):
    @staticmethod
    def _mapper_func(report_path, parsed_args):
        service = StatsService(report_path)

        df = service.read_sql_report(
            "cuda_gpu_mem_time_sum", "CudaGpuMemTimeSum", parsed_args
        )
        if df is None:
            return None

        if df.empty:
            logger.info(
                f"{report_path} was successfully processed, but no data was found."
            )
            return None

        df["File"] = Path(report_path).stem
        return df

    @log.time("Mapper")
    def mapper_func(self, context):
        return context.wait(
            context.map(
                self._mapper_func,
                self._parsed_args.input,
                parsed_args=self._parsed_args,
            )
        )

    def aggregate_stats(self, df):
        grouped = df.groupby(df.index.name)

        d = OrderedDict()
        total_time = grouped["Total Time"].sum()
        d["Time"] = total_time / total_time.sum() * 100
        d["Total Time"] = total_time
        d["Count"] = grouped["Count"].sum()
        d["Avg"] = d["Total Time"] / d["Count"]
        d["Q1"] = grouped["Q1"].min()
        d["Med"] = grouped["Med"].median()
        d["Q3"] = grouped["Q3"].max()
        d["Min"] = grouped["Min"].min()
        d["Max"] = grouped["Max"].max()
        d["StdDev"] = grouped.apply(lambda x: helpers.stddev(x, d, "Count"))
        min_value = grouped["Min"].min()
        d["Min File"] = grouped.apply(
            lambda x: ", ".join(x.loc[x["Min"] == min_value.loc[x.name], "File"])
        )
        max_value = grouped["Max"].max()
        d["Max File"] = grouped.apply(
            lambda x: ", ".join(x.loc[x["Max"] == max_value.loc[x.name], "File"])
        )

        aggregated_df = pd.concat(d.values(), axis=1, keys=d.keys()).round(1)
        return aggregated_df.sort_values(by=["Total Time"], ascending=False)

    @log.time("Reducer")
    def reducer_func(self, dfs):
        dfs = helpers.filter_none(dfs)
        df = pd.concat(dfs)

        # Remove any tags or hidden columns that are for internal use.
        df.columns = df.columns.str.replace("(:).*", "", regex=True)
        df.columns = df.columns.str.lstrip("_")

        df = df.set_index("Operation")
        df = df[
            [
                "Time",
                "Total Time",
                "Count",
                "Avg",
                "Q1",
                "Med",
                "Q3",
                "Min",
                "Max",
                "StdDev",
                "File",
            ]
        ]

        per_rank_df = df.sort_values(by=["File", "Total Time"], ascending=[True, False])
        all_ranks_df = self.aggregate_stats(df)

        per_rank_df.to_parquet(self.add_output_file("rank_stats.parquet"))
        all_ranks_df.to_parquet(self.add_output_file("all_stats.parquet"))

        if self._parsed_args.csv:
            per_rank_df.to_csv(self.add_output_file("rank_stats.csv"))
            all_ranks_df.to_csv(self.add_output_file("all_stats.csv"))

    def save_notebook(self):
        self.create_notebook("stats.ipynb")
        self.add_notebook_helper_file("nsys_display.py")

    def save_analysis_file(self):
        self._analysis_dict.update(
            {
                "EndTime": str(datetime.now()),
                "InputFiles": self._parsed_args.input,
                "Outputs": self._output_files,
            }
        )
        self.create_analysis_file()

    def run(self, context):
        super().run(context)

        mapper_res = self.mapper_func(context)
        self.reducer_func(mapper_res)

        self.save_notebook()
        self.save_analysis_file()

    @classmethod
    def get_argument_parser(cls):
        parser = super().get_argument_parser()

        parser.add_recipe_argument(Option.INPUT)
        parser.add_recipe_argument(Option.OUTPUT)
        parser.add_recipe_argument(Option.FORCE_OVERWRITE)
        parser.add_recipe_argument(Option.START)
        parser.add_recipe_argument(Option.END)
        parser.add_recipe_argument(Option.CSV)

        return parser
