# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.

import pandas as pd
import json

from pathlib import Path
from datetime import datetime

from nsys_recipe import log, data_service
from nsys_recipe.data_service import DataService
from nsys_recipe.lib import heatmap, helpers, recipe
from nsys_recipe.lib.args import Option
from nsys_recipe.lib.collective_loader import ProfileInfo
from nsys_recipe.log import logger


class GpuMetricUtilMap(recipe.Recipe):
    @staticmethod
    def _mapper_func(report_path, bin_list, min_session, parsed_args):
        service = DataService(report_path)

        table_column_dict = {
            "ANALYSIS_DETAILS": ["duration"],
            "GENERIC_EVENTS": ["rawTimestamp", "typeId", "data"],
            "TARGET_INFO_SESSION_START_TIME": ["utcEpochNs"],
        }

        df_dict = service.read_tables(table_column_dict)
        if df_dict is None:
            return None

        df = df_dict["GENERIC_EVENTS"]
        session_offset, profile_duration = data_service.compute_session_duration(
            df_dict["ANALYSIS_DETAILS"],
            df_dict["TARGET_INFO_SESSION_START_TIME"],
            min_session,
        )

        data_service.filter_by_time_range(df, parsed_args.start, parsed_args.end)
        data_service.apply_time_offset(df, session_offset)

        if df.empty:
            logger.info(
                f"{report_path} was successfully processed, but no data was found."
            )
            return None

        # Transform the 'data' column from JSON strings into individual columns.
        df["data"] = df["data"].apply(json.loads)
        df_json = pd.json_normalize(df["data"])
        df = pd.concat([df, df_json], axis=1)

        # Filter the DataFrame based on the profiling duration.
        # TODO(DTSP-15842): Use 'timestamp' instead of 'rawTimestamp'.
        df = df[
            (df["rawTimestamp"] >= session_offset)
            & (df["rawTimestamp"] <= profile_duration)
        ]

        # Extract GPU ID from type ID.
        df["GPU"] = df["typeId"] & 0xFF
        # Arrange the GPU metric sample points into their respective duration bins.
        df["Duration"] = pd.cut(df["rawTimestamp"], bin_list, labels=bin_list[:-1])

        # Convert the metrics to int type.
        metric_cols = ["SM Active", "SM Issue", "Tensor Active"]
        df[metric_cols] = df[metric_cols].astype(int)

        # Calculate the mean for the specified metrics per GPU and bin.
        df = (
            df.groupby(["GPU", "Duration"], observed=True)[metric_cols]
            .mean()
            .round(1)
            .reset_index()
        )
        df["File"] = Path(report_path).stem

        return df[["Duration", *metric_cols, "GPU", "File"]]

    @log.time("Mapper")
    def mapper_func(self, context, filtered_dir, bin_list, min_session):
        return context.wait(
            context.map(
                self._mapper_func,
                filtered_dir,
                bin_list=bin_list,
                min_session=min_session,
                parsed_args=self._parsed_args,
            )
        )

    @log.time("Reducer")
    def reducer_func(self, dfs):
        dfs = helpers.filter_none(dfs)
        df = pd.concat(dfs)

        df.to_parquet(self.add_output_file("analysis.parquet"))

    def save_notebook(self):
        self.create_notebook(
            "heatmap.ipynb", replace_dict={"REPLACE_BIN": self._parsed_args.bins}
        )
        self.add_notebook_helper_file("nsys_display.py")

    def save_analysis_file(self):
        self._analysis_dict.update(
            {
                "EndTime": str(datetime.now()),
                "InputFiles": self._parsed_args.input,
                "Outputs": self._output_files,
            }
        )
        self.create_analysis_file()

    def run(self, context):
        super().run(context)

        max_duration, min_session, filtered_dir = ProfileInfo.get_profile_info(
            context, self._parsed_args.input
        )
        bin_size = heatmap.get_bin_size(self._parsed_args.bins, max_duration)
        bin_list = heatmap.generate_bin_list(
            self._parsed_args.bins, bin_size, include_last=True
        )

        mapper_res = self.mapper_func(context, filtered_dir, bin_list, min_session)
        self.reducer_func(mapper_res)

        self.save_notebook()
        self.save_analysis_file()

    @classmethod
    def get_argument_parser(cls):
        parser = super().get_argument_parser()

        parser.add_recipe_argument(Option.INPUT)
        parser.add_recipe_argument(Option.OUTPUT)
        parser.add_recipe_argument(Option.FORCE_OVERWRITE)
        parser.add_recipe_argument(Option.START)
        parser.add_recipe_argument(Option.END)
        parser.add_recipe_argument(Option.BINS)

        return parser
