# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.

import pandas as pd

from pathlib import Path
from datetime import datetime

from nsys_recipe import log, data_service
from nsys_recipe.data_service import DataService
from nsys_recipe.lib import helpers, recipe, heatmap
from nsys_recipe.lib.args import Option
from nsys_recipe.lib.collective_loader import ProfileInfo
from nsys_recipe.log import logger


class MpiGpuTimeUtilMap(recipe.Recipe):
    @staticmethod
    def _mapper_func(report_path, bin_size, min_session, parsed_args):
        service = DataService(report_path)

        table_column_dict = {
            "ANALYSIS_DETAILS": ["duration"],
            "TARGET_INFO_SESSION_START_TIME": ["utcEpochNs"],
            "CUPTI_ACTIVITY_KIND_RUNTIME": [
                "correlationId",
                "globalTid",
                "start",
                "end",
            ],
            "CUPTI_ACTIVITY_KIND_KERNEL": [
                "correlationId",
                "globalPid",
                "start",
                "end",
            ],
            "MPI_P2P_EVENTS": ["globalTid", "start", "end"],
            "MPI_START_WAIT_EVENTS": ["globalTid", "start", "end"],
            "MPI_OTHER_EVENTS": ["globalTid", "start", "end"],
            "MPI_COLLECTIVES_EVENTS": ["globalTid", "start", "end"],
        }

        df_dict = service.read_tables(table_column_dict)
        if df_dict is None:
            return None

        mpi_p2p_df = df_dict["MPI_P2P_EVENTS"]
        mpi_start_wait_df = df_dict["MPI_START_WAIT_EVENTS"]
        mpi_other_df = df_dict["MPI_OTHER_EVENTS"]
        mpi_collectives_df = df_dict["MPI_COLLECTIVES_EVENTS"]

        mpi_dfs = [mpi_p2p_df, mpi_start_wait_df, mpi_other_df, mpi_collectives_df]
        kernel_df = df_dict["CUPTI_ACTIVITY_KIND_KERNEL"]
        dfs = [*mpi_dfs, kernel_df]

        session_offset, profile_duration = data_service.compute_session_duration(
            df_dict["ANALYSIS_DETAILS"],
            df_dict["TARGET_INFO_SESSION_START_TIME"],
            min_session,
        )

        data_service.filter_by_time_range(dfs, parsed_args.start, parsed_args.end)
        data_service.apply_time_offset(dfs, session_offset)

        mpi_df = pd.concat(mpi_dfs).sort_values("start").reset_index()
        if mpi_df.empty or kernel_df.empty:
            logger.info(
                f"{report_path} was successfully processed, but no data was found."
            )
            return None

        kernel_df["pid"] = data_service.extract_pid(kernel_df["globalPid"])
        mpi_df["pid"] = data_service.extract_pid(mpi_df["globalTid"])

        kernel_grouped = kernel_df.groupby("pid")
        mpi_grouped = mpi_df.groupby("pid")

        unique_pids = set(kernel_grouped.groups.keys()) | set(mpi_grouped.groups.keys())
        results = []

        for pid in unique_pids:
            group_dfs = {}
            bin_pcts = {}

            # Initialize the bins with zero values.
            zero_bins = heatmap.get_zero_bin_pcts(
                bin_size, parsed_args.bins, profile_duration, session_offset
            )
            bin_pcts = {key: zero_bins for key in ("Kernel", "MPI", "Overlap")}

            groups_info = [("Kernel", kernel_grouped), ("MPI", mpi_grouped)]
            for name, grouped in groups_info:
                if pid in grouped.groups:
                    group_dfs[name] = heatmap.consolidate_ranges(grouped.get_group(pid))

            # Calculate the overlap only if 'Kernel' and 'MPI' both exist for
            # the given PID.
            if "Kernel" in group_dfs and "MPI" in group_dfs:
                overlap_range_df = heatmap.calculate_overlapping_ranges(
                    group_dfs["Kernel"], group_dfs["MPI"]
                )
                group_dfs["Overlap"] = overlap_range_df

            # Calculate bin percentages for each dataframe.
            for name, group_df in group_dfs.items():
                bin_pcts[name] = heatmap.calculate_bin_pcts(
                    group_df,
                    bin_size,
                    parsed_args.bins,
                    profile_duration,
                    session_offset,
                )

            data = {
                "Duration": heatmap.generate_bin_list(parsed_args.bins, bin_size),
                **bin_pcts,
                "PID": pid,
            }
            results.append(pd.DataFrame(data))

        pct_df = pd.concat(results, ignore_index=True)
        pct_df["File"] = Path(report_path).stem

        return pct_df

    @log.time("Mapper")
    def mapper_func(self, context, filtered_dir, bin_size, min_session):
        return context.wait(
            context.map(
                self._mapper_func,
                filtered_dir,
                bin_size=bin_size,
                min_session=min_session,
                parsed_args=self._parsed_args,
            )
        )

    @log.time("Reducer")
    def reducer_func(self, dfs):
        dfs = helpers.filter_none(dfs)
        df = pd.concat(dfs)

        df.to_parquet(self.add_output_file("analysis.parquet"))

    def save_notebook(self):
        self.create_notebook(
            "heatmap.ipynb", replace_dict={"REPLACE_BIN": self._parsed_args.bins}
        )
        self.add_notebook_helper_file("nsys_display.py")

    def save_analysis_file(self):
        self._analysis_dict.update(
            {
                "EndTime": str(datetime.now()),
                "InputFiles": self._parsed_args.input,
                "Outputs": self._output_files,
            }
        )
        self.create_analysis_file()

    def run(self, context):
        super().run(context)

        max_duration, min_session, filtered_dir = ProfileInfo.get_profile_info(
            context, self._parsed_args.input
        )
        bin_size = heatmap.get_bin_size(self._parsed_args.bins, max_duration)

        mapper_res = self.mapper_func(context, filtered_dir, bin_size, min_session)
        self.reducer_func(mapper_res)

        self.save_notebook()
        self.save_analysis_file()

    @classmethod
    def get_argument_parser(cls):
        parser = super().get_argument_parser()

        parser.add_recipe_argument(Option.INPUT)
        parser.add_recipe_argument(Option.OUTPUT)
        parser.add_recipe_argument(Option.FORCE_OVERWRITE)
        parser.add_recipe_argument(Option.START)
        parser.add_recipe_argument(Option.END)
        parser.add_recipe_argument(Option.BINS)
        return parser
