# SPDX-FileCopyrightText: Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.

import pandas as pd

from datetime import datetime
from pathlib import Path

from nsys_recipe import log, data_service
from nsys_recipe.data_service import DataService
from nsys_recipe.lib import helpers, recipe, summary
from nsys_recipe.lib.args import Option
from nsys_recipe.log import logger


class MpiSum(recipe.Recipe):
    @staticmethod
    def _mapper_func(report_path, parsed_args):
        service = DataService(report_path)

        table_column_dict = {
            "StringIds": None,
            "MPI_P2P_EVENTS": ["start", "end", "textId"],
            "MPI_START_WAIT_EVENTS": ["start", "end", "textId"],
            "MPI_OTHER_EVENTS": ["start", "end", "textId"],
            "MPI_COLLECTIVES_EVENTS": ["start", "end", "textId"],
        }

        df_dict = service.read_tables(table_column_dict)
        if df_dict is None:
            return None

        mpi_p2p_df = df_dict["MPI_P2P_EVENTS"]
        mpi_start_wait_df = df_dict["MPI_START_WAIT_EVENTS"]
        mpi_other_df = df_dict["MPI_OTHER_EVENTS"]
        mpi_collectives_df = df_dict["MPI_COLLECTIVES_EVENTS"]

        mpi_dfs = [mpi_p2p_df, mpi_start_wait_df, mpi_other_df, mpi_collectives_df]
        data_service.filter_by_time_range(mpi_dfs, parsed_args.start, parsed_args.end)

        mpi_df = pd.concat(mpi_dfs)
        if mpi_df.empty:
            logger.info(
                f"{report_path} was successfully processed, but no data was found."
            )
            return None

        mpi_df = data_service.replace_id_with_value(
            mpi_df, df_dict["StringIds"], "textId"
        )
        mpi_df = mpi_df.rename(columns={"textId": "text"})

        mpi_df["duration"] = mpi_df["end"] - mpi_df["start"]
        mpi_df = mpi_df[mpi_df["duration"] > 0]
        range_gdf = mpi_df[["text", "duration"]].groupby("text")

        filename = Path(report_path).stem
        stats_df = summary.describe_duration(range_gdf["duration"])

        return summary.StatsInfo(filename, stats_df)

    @log.time("Mapper")
    def mapper_func(self, context):
        return context.wait(
            context.map(
                self._mapper_func,
                self._parsed_args.input,
                parsed_args=self._parsed_args,
            )
        )

    @log.time("Reducer")
    def reducer_func(self, mapper_res):
        stats_infos = helpers.filter_none(mapper_res)
        stats_infos = sorted(stats_infos, key=lambda x: x.filename)

        filenames, stats_dfs = zip(*stats_infos)

        rank_file_df = pd.DataFrame(filenames, columns=["File"])
        rank_file_df.to_parquet(self.add_output_file("files.parquet"), index=False)

        all_stats_df = summary.aggregate_stats_dfs(stats_dfs)
        all_stats_df.to_parquet(self.add_output_file("all_stats.parquet"))

        rank_stats_dfs = [
            df.assign(Rank=rank, File=filenames[rank])
            for rank, df in enumerate(stats_dfs)
        ]
        rank_stats = pd.concat(rank_stats_dfs)
        rank_stats.to_parquet(self.add_output_file("rank_stats.parquet"))

        if self._parsed_args.csv:
            all_stats_df.to_csv(self.add_output_file("all_stats.csv"))
            rank_stats.to_csv(self.add_output_file("rank_stats.csv"))

    def save_notebook(self):
        self.create_notebook("stats.ipynb")
        self.add_notebook_helper_file("nsys_pres.py")

    def save_analysis_file(self):
        self._analysis_dict.update(
            {
                "EndTime": str(datetime.now()),
                "InputFiles": self._parsed_args.input,
                "Outputs": self._output_files,
            }
        )
        self.create_analysis_file()

    def run(self, context):
        super().run(context)

        mapper_res = self.mapper_func(context)
        self.reducer_func(mapper_res)

        self.save_notebook()
        self.save_analysis_file()

    @classmethod
    def get_argument_parser(cls):
        parser = super().get_argument_parser()

        parser.add_recipe_argument(Option.INPUT)
        parser.add_recipe_argument(Option.OUTPUT)
        parser.add_recipe_argument(Option.FORCE_OVERWRITE)
        parser.add_recipe_argument(Option.START)
        parser.add_recipe_argument(Option.END)
        parser.add_recipe_argument(Option.CSV)

        return parser
