# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.

import pandas as pd

from pathlib import Path
from datetime import datetime

from nsys_recipe import log, data_service
from nsys_recipe.data_service import DataService
from nsys_recipe.lib import helpers, nvtx, recipe, heatmap
from nsys_recipe.lib.args import Option
from nsys_recipe.lib.collective_loader import ProfileInfo
from nsys_recipe.log import logger


class NcclGpuTimeUtilMap(recipe.Recipe):
    @staticmethod
    def _mapper_func(report_path, bin_size, min_session, parsed_args):
        service = DataService(report_path)

        table_column_dict = {
            "ANALYSIS_DETAILS": ["duration"],
            "TARGET_INFO_SESSION_START_TIME": ["utcEpochNs"],
            "StringIds": None,
            "NVTX_EVENTS": [
                "text",
                "start",
                "end",
                "textId",
                "globalTid",
                "endGlobalTid",
                "domainId",
                "eventType",
            ],
            "CUPTI_ACTIVITY_KIND_RUNTIME": [
                "correlationId",
                "globalTid",
                "start",
                "end",
            ],
            "CUPTI_ACTIVITY_KIND_KERNEL": [
                "correlationId",
                "globalPid",
                "start",
                "end",
                "deviceId",
            ],
        }

        df_dict = service.read_tables(table_column_dict)
        if df_dict is None:
            return None

        kernel_df = df_dict["CUPTI_ACTIVITY_KIND_KERNEL"]
        session_offset, profile_duration = data_service.compute_session_duration(
            df_dict["ANALYSIS_DETAILS"],
            df_dict["TARGET_INFO_SESSION_START_TIME"],
            min_session,
        )

        data_service.filter_by_time_range(kernel_df, parsed_args.start, parsed_args.end)
        data_service.apply_time_offset(kernel_df, session_offset)

        cuda_df = data_service.combine_api_gpu_dfs(
            df_dict["CUPTI_ACTIVITY_KIND_RUNTIME"], kernel_df
        )
        nccl_df = data_service.get_domain_range_df(df_dict["NVTX_EVENTS"], "NCCL")

        if nccl_df.empty or cuda_df.empty:
            logger.info(
                f"{report_path} was successfully processed, but no data was found."
            )
            return None

        # Project NCCL kernels onto GPU.
        nccl_df = data_service.combine_text_fields(nccl_df, df_dict["StringIds"])
        nccl_gpu_df = nvtx.project_nvtx_onto_gpu(nccl_df, cuda_df)
        if nccl_gpu_df.empty:
            logger.info(
                f"{report_path} does not contain any NCCL data that can be projected onto the GPU."
            )
            return None

        # Merge the kernel dataframe with the projected NCCL dataframe.
        # If a kernel exists in both dataframes, it is an NCCL kernel.
        # Otherwise, it is a compute kernel.
        merged_df = kernel_df.merge(
            nccl_gpu_df, on=["start", "end"], how="left", indicator="merged"
        )
        merged_df["pid"] = data_service.extract_pid(merged_df["globalPid"])

        kernel_grouped = merged_df.groupby(["pid", "deviceId"])
        results = []

        for (pid, deviceId), group_df in kernel_grouped:
            type_dfs = {
                "All": group_df,
                "NCCL": group_df[group_df["merged"] == "both"],
                "Compute": group_df[group_df["merged"] == "left_only"],
            }
            bin_pcts = {}

            for name, type_df in type_dfs.items():
                if not parsed_args.cumulative:
                    type_df = heatmap.consolidate_ranges(type_df)

                bin_pcts[name] = heatmap.calculate_bin_pcts(
                    type_df,
                    bin_size,
                    parsed_args.bins,
                    profile_duration,
                    session_offset,
                )

            data = {
                "Duration": heatmap.generate_bin_list(parsed_args.bins, bin_size),
                **bin_pcts,
                "PID": pid,
                "Device ID": deviceId,
            }
            results.append(pd.DataFrame(data))

        pct_df = pd.concat(results, ignore_index=True)
        pct_df["File"] = Path(report_path).stem

        return pct_df

    @log.time("Mapper")
    def mapper_func(self, context, filtered_dir, bin_size, min_session):
        return context.wait(
            context.map(
                self._mapper_func,
                filtered_dir,
                bin_size=bin_size,
                min_session=min_session,
                parsed_args=self._parsed_args,
            )
        )

    @log.time("Reducer")
    def reducer_func(self, dfs):
        dfs = helpers.filter_none(dfs)
        df = pd.concat(dfs)

        df.to_parquet(self.add_output_file("analysis.parquet"))

    def save_notebook(self):
        self.create_notebook(
            "heatmap.ipynb", replace_dict={"REPLACE_BIN": self._parsed_args.bins}
        )
        self.add_notebook_helper_file("nsys_display.py")

    def save_analysis_file(self):
        self._analysis_dict.update(
            {
                "EndTime": str(datetime.now()),
                "InputFiles": self._parsed_args.input,
                "Outputs": self._output_files,
            }
        )
        self.create_analysis_file()

    def run(self, context):
        super().run(context)

        max_duration, min_session, filtered_dir = ProfileInfo.get_profile_info(
            context, self._parsed_args.input
        )
        bin_size = heatmap.get_bin_size(self._parsed_args.bins, max_duration)

        mapper_res = self.mapper_func(context, filtered_dir, bin_size, min_session)
        self.reducer_func(mapper_res)

        self.save_notebook()
        self.save_analysis_file()

    @classmethod
    def get_argument_parser(cls):
        parser = super().get_argument_parser()

        parser.add_recipe_argument(Option.INPUT)
        parser.add_recipe_argument(Option.OUTPUT)
        parser.add_recipe_argument(Option.FORCE_OVERWRITE)
        parser.add_recipe_argument(Option.START)
        parser.add_recipe_argument(Option.END)
        parser.add_recipe_argument(Option.BINS)
        parser.add_recipe_argument(
            "--cumulative",
            action="store_true",
            help="Add up the utilization of all kernels within each bin."
            " The utilization can exceed 100%% if multiple kernels are running concurrently",
        )
        return parser
