# SPDX-FileCopyrightText: Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.

import pandas as pd

from pathlib import Path
from datetime import datetime

from nsys_recipe import log, data_service
from nsys_recipe.data_service import DataService
from nsys_recipe.lib import helpers, nvtx, recipe, heatmap
from nsys_recipe.lib.args import Option
from nsys_recipe.lib.collective_loader import ProfileInfo
from nsys_recipe.log import logger


class NicMetricMap(recipe.Recipe):
    @staticmethod
    def _mapper_func(report_path, bin_size, min_session, parsed_args):
        service = DataService(report_path)

        table_column_dict = {
            "ANALYSIS_DETAILS": ["duration"],
            "TARGET_INFO_SESSION_START_TIME": ["utcEpochNs"],
            "StringIds": None,
            "TARGET_INFO_NETWORK_METRICS": ["metricsListId", "metricsIdx", "name"],
            "NIC_ID_MAP": ["globalId", "nicId"],
            "NET_NIC_METRIC": [
                "start",
                "end",
                "globalId",
                "metricsListId",
                "metricsIdx",
                "value",
            ],
        }

        df_dict = service.read_tables(table_column_dict)
        if df_dict is None:
            return None

        nic_metric_df = df_dict["NET_NIC_METRIC"].join(
            df_dict["NIC_ID_MAP"].set_index("globalId"), on="globalId"
        )
        session_offset, profile_duration = data_service.compute_session_duration(
            df_dict["ANALYSIS_DETAILS"],
            df_dict["TARGET_INFO_SESSION_START_TIME"],
            min_session,
        )

        data_service.filter_by_time_range(
            nic_metric_df, parsed_args.start, parsed_args.end
        )
        data_service.apply_time_offset(nic_metric_df, session_offset)

        if nic_metric_df.empty:
            logger.info(
                f"{report_path} was successfully processed, but no data was found."
            )
            return None

        nic_metric_grouped = nic_metric_df.groupby(["globalId", "nicId"])
        results = []
        network_metric_info_df = df_dict["TARGET_INFO_NETWORK_METRICS"]

        for (globalId, nicId), group_df in nic_metric_grouped:
            bytes_received_id = network_metric_info_df.loc[
                network_metric_info_df["name"] == "IB: Bytes sent", "metricsIdx"
            ].iloc[0]
            df_recv = group_df.loc[
                group_df["metricsIdx"] == bytes_received_id
            ].reset_index(drop=True)
            bytes_sent_id = network_metric_info_df.loc[
                network_metric_info_df["name"] == "IB: Bytes received", "metricsIdx"
            ].iloc[0]
            df_sent = group_df.loc[group_df["metricsIdx"] == bytes_sent_id].reset_index(
                drop=True
            )
            congestion_id = network_metric_info_df.loc[
                network_metric_info_df["name"] == "IB: Send waits", "metricsIdx"
            ].iloc[0]
            type_dfs = {
                "Bytes Recv": df_recv,
                "Bytes Sent": df_sent,
                "Congestion": group_df.loc[
                    group_df["metricsIdx"] == congestion_id
                ].reset_index(drop=True),
            }
            bin_pcts = {}

            for name, type_df in type_dfs.items():
                type_df = type_df.loc[type_df["value"] > 0].reset_index(drop=True)
                bin_pcts[name] = heatmap.calculate_bin_pcts(
                    type_df,
                    bin_size,
                    parsed_args.bins,
                    profile_duration,
                    session_offset,
                    value_key="value",
                )

            data = {
                "Duration": heatmap.generate_bin_list(parsed_args.bins, bin_size),
                **bin_pcts,
                "GlobalId": globalId,
                "NIC ID": nicId,
            }
            results.append(pd.DataFrame(data))

        pct_df = pd.concat(results, ignore_index=True)
        pct_df["File"] = Path(report_path).stem

        return pct_df

    @log.time("Mapper")
    def mapper_func(self, context, filtered_dir, bin_size, min_session):
        return context.wait(
            context.map(
                self._mapper_func,
                filtered_dir,
                bin_size=bin_size,
                min_session=min_session,
                parsed_args=self._parsed_args,
            )
        )

    @log.time("Reducer")
    def reducer_func(self, dfs):
        dfs = helpers.filter_none(dfs)
        df = pd.concat(dfs)

        df.to_parquet(self.add_output_file("analysis.parquet"))

    def save_notebook(self):
        self.create_notebook(
            "heatmap.ipynb", replace_dict={"REPLACE_BIN": self._parsed_args.bins}
        )
        self.add_notebook_helper_file("nsys_display.py")

    def save_analysis_file(self):
        self._analysis_dict.update(
            {
                "EndTime": str(datetime.now()),
                "InputFiles": self._parsed_args.input,
                "Outputs": self._output_files,
            }
        )
        self.create_analysis_file()

    def run(self, context):
        super().run(context)

        max_duration, min_session, filtered_dir = ProfileInfo.get_profile_info(
            context, self._parsed_args.input
        )
        bin_size = heatmap.get_bin_size(self._parsed_args.bins, max_duration)

        mapper_res = self.mapper_func(context, filtered_dir, bin_size, min_session)
        self.reducer_func(mapper_res)

        self.save_notebook()
        self.save_analysis_file()

    @classmethod
    def get_argument_parser(cls):
        parser = super().get_argument_parser()

        parser.add_recipe_argument(Option.INPUT)
        parser.add_recipe_argument(Option.OUTPUT)
        parser.add_recipe_argument(Option.FORCE_OVERWRITE)
        parser.add_recipe_argument(Option.START)
        parser.add_recipe_argument(Option.END)
        parser.add_recipe_argument(Option.BINS)

        return parser
