# SPDX-FileCopyrightText: Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.

import pandas as pd

from datetime import datetime
from pathlib import Path

from nsys_recipe import log, data_service
from nsys_recipe.data_service import DataService
from nsys_recipe.lib import helpers, nvtx, recipe, summary
from nsys_recipe.lib.args import Option
from nsys_recipe.log import logger


class NvtxGpuProjSum(recipe.Recipe):
    @staticmethod
    def _mapper_func(report_path, parsed_args):
        service = DataService(report_path)

        name_column = "text"
        table_column_dict = {
            "StringIds": None,
            "NVTX_EVENTS": [
                name_column,
                "start",
                "end",
                "textId",
                "globalTid",
                "endGlobalTid",
            ],
            "CUPTI_ACTIVITY_KIND_RUNTIME": [
                "correlationId",
                "globalTid",
                "start",
                "end",
            ],
            "CUPTI_ACTIVITY_KIND_KERNEL": [
                "correlationId",
                "globalPid",
                "start",
                "end",
            ],
            "CUPTI_ACTIVITY_KIND_MEMCPY": [
                "correlationId",
                "globalPid",
                "start",
                "end",
            ],
            "CUPTI_ACTIVITY_KIND_MEMSET": [
                "correlationId",
                "globalPid",
                "start",
                "end",
            ],
        }

        df_dict = service.read_tables(table_column_dict)
        if df_dict is None:
            return None

        nvtx_df = df_dict["NVTX_EVENTS"]
        kernel_df = df_dict["CUPTI_ACTIVITY_KIND_KERNEL"]
        memcpy_df = df_dict["CUPTI_ACTIVITY_KIND_MEMCPY"]
        memset_df = df_dict["CUPTI_ACTIVITY_KIND_MEMSET"]

        gpu_dfs = [kernel_df, memcpy_df, memset_df]
        data_service.filter_by_time_range(gpu_dfs, parsed_args.start, parsed_args.end)

        gpu_df = pd.concat(gpu_dfs, ignore_index=True)
        cuda_df = data_service.combine_api_gpu_dfs(
            df_dict["CUPTI_ACTIVITY_KIND_RUNTIME"], gpu_df
        )

        if cuda_df.empty or nvtx_df.empty:
            logger.info(
                f"{report_path} was successfully processed, but no data was found."
            )
            return None

        nvtx_df = data_service.combine_text_fields(nvtx_df, df_dict["StringIds"])
        nvtx_gpu_df = nvtx.project_nvtx_onto_gpu(nvtx_df, cuda_df)
        if nvtx_gpu_df.empty:
            logger.info(
                f"{report_path} does not contain any NVTX data that can be projected onto the GPU."
            )
            return None

        nvtx_gpu_df["duration"] = nvtx_gpu_df["end"] - nvtx_gpu_df["start"]
        nvtx_gpu_df = nvtx_gpu_df[nvtx_gpu_df["duration"] > 0]
        range_gdf = nvtx_gpu_df[["text", "duration"]].groupby("text")

        filename = Path(report_path).stem
        stats_df = summary.describe_duration(range_gdf["duration"])

        return summary.StatsInfo(filename, stats_df)

    @log.time("Mapper")
    def mapper_func(self, context):
        return context.wait(
            context.map(
                self._mapper_func,
                self._parsed_args.input,
                parsed_args=self._parsed_args,
            )
        )

    @log.time("Reducer")
    def reducer_func(self, mapper_res):
        stats_infos = helpers.filter_none(mapper_res)
        stats_infos = sorted(stats_infos, key=lambda x: x.filename)

        filenames, stats_dfs = zip(*stats_infos)

        rank_file_df = pd.DataFrame(filenames, columns=["File"])
        rank_file_df.to_parquet(self.add_output_file("files.parquet"), index=False)

        all_stats_df = summary.aggregate_stats_dfs(stats_dfs)
        all_stats_df.to_parquet(self.add_output_file("all_stats.parquet"))

        rank_stats_dfs = [
            df.assign(Rank=rank, File=filenames[rank])
            for rank, df in enumerate(stats_dfs)
        ]
        rank_stats = pd.concat(rank_stats_dfs)
        rank_stats.to_parquet(self.add_output_file("rank_stats.parquet"))

        if self._parsed_args.csv:
            all_stats_df.to_csv(self.add_output_file("all_stats.csv"))
            rank_stats.to_csv(self.add_output_file("rank_stats.csv"))

    def save_notebook(self):
        self.create_notebook("stats.ipynb")
        self.add_notebook_helper_file("nsys_pres.py")

    def save_analysis_file(self):
        self._analysis_dict.update(
            {
                "EndTime": str(datetime.now()),
                "InputFiles": self._parsed_args.input,
                "Outputs": self._output_files,
            }
        )
        self.create_analysis_file()

    def run(self, context):
        super().run(context)

        mapper_res = self.mapper_func(context)
        self.reducer_func(mapper_res)

        self.save_notebook()
        self.save_analysis_file()

    @classmethod
    def get_argument_parser(cls):
        parser = super().get_argument_parser()

        parser.add_recipe_argument(Option.INPUT)
        parser.add_recipe_argument(Option.OUTPUT)
        parser.add_recipe_argument(Option.FORCE_OVERWRITE)
        parser.add_recipe_argument(Option.START)
        parser.add_recipe_argument(Option.END)
        parser.add_recipe_argument(Option.CSV)

        return parser
